@@ -151,6 +151,7 @@ from warnings import warn
 from i2 import double_up_as_factory
 from i2.signatures import (
     call_somewhat_forgivingly,
+    call_forgivingly,
     Parameter,
     empty,
     Sig,
@@ -158,6 +159,7 @@ from i2.signatures import (
 )
 from meshed.base import (
     FuncNode,
+    ch_func_node_attrs,
     validate_that_func_node_names_are_sane,
     _mk_func_nodes,
     _func_nodes_to_graph_dict,
@@ -190,6 +192,16 @@ from meshed.itools import (
     ancestors,
 )
 
+dflt_configs = dict(
+    fnode_shape='box',
+    vnode_shape='none',
+    display_all_arguments=True,
+    edge_kind='to_args_on_edge',
+    input_node=True,
+    output_node='output',
+    func_display=True,
+)
+
 
 def order_subset_from_list(items, sublist):
     assert set(sublist).issubset(set(items)), f'{sublist} is not contained in {items}'
@@ -233,6 +245,24 @@ def mk_func_name(func, exclude_names=()):
     return find_first_free_name(name, exclude_names)
 
 
+def mk_nodes_names_unique(nodes, exclude_names=()):
+    names = [node.name for node in nodes]
+
+    def gen():
+        _exclude_names = exclude_names
+        for name in names:
+            if name not in _exclude_names:
+                yield name
+            else:
+                found_name = find_first_free_name(
+                    f'{func_name}__{name}', _exclude_names
+                )
+                yield found_name
+                _exclude_names = _exclude_names + (found_name,)
+
+    return list(gen())
+
+
 def arg_names(func, func_name, exclude_names=()):
     names = Sig(func).names
 
@@ -397,6 +427,12 @@ def change_value_on_cond(d, cond, func):
     return d
 
 
+def dflt_debugger_feedback(func_node, scope, output, step):
+    print(f'{step} --------------------------------------------------------------')
+    print(f'\t{func_node=}\n\t{scope=}')
+    return output
+
+
 # TODO: caching last scope isn't really the DAG's direct concern -- it's a debugging
 #  concern. Perhaps a more general form would be to define a cache factory defaulting
 #  to a dict, but that could be a "dict" that logs writes (even to an attribute of self)
@@ -480,29 +516,41 @@ class DAG:
         # But we want leafs in topological order
         self.leafs = tuple([name for name in self.nodes if name in leafs])
         self.last_scope = None
-        if self.name is not None:
-            self.__name__ = self.name
+        self.__name__ = self.name or 'DAG'
 
         self.bindings_cleaner()
 
     def bindings_cleaner(self):
-        funcnodes_names = [node.name for node in self.func_nodes]
+        old_nodes = self.func_nodes
+        old_names = [node.name for node in old_nodes]
+        funcnodes_names = mk_nodes_names_unique(old_nodes)
+        dict_renamer = {
+            old_name: new_name for old_name, new_name in zip(old_names, funcnodes_names)
+        }
+        self.func_nodes = ch_names(old_nodes, renamer=dict_renamer)
         func = lambda v: self._func_node_for[v].out
         cond = lambda k, v: v in funcnodes_names
-        for node in self.func_nodes:
+        for node, name in zip(self.func_nodes, funcnodes_names):
+            node.name = name
             node.bind = change_value_on_cond(node.bind, cond, func)
 
     def __call__(self, *args, **kwargs):
         return self._call(*args, **kwargs)
 
-    def _call(self, *args, **kwargs):
-        # Get a dict of {argname: argval} pairs from positional and keyword arguments
-        # How positionals are resolved is determined by sels.sig
-        # The result is the initial ``scope`` the func nodes will both read from
-        # to get their arguments, and write their outputs to.
-        scope = self.__signature__.kwargs_from_args_and_kwargs(
+    def _get_kwargs(self, *args, **kwargs):
+        """
+        Get a dict of {argname: argval} pairs from positional and keyword arguments.
+        """
+        return self.__signature__.kwargs_from_args_and_kwargs(
             args, kwargs, apply_defaults=True
         )
+
+    def _call(self, *args, **kwargs):
+        # Get a dict of {argname: argval} pairs from positional and keyword arguments
+        # How positionals are resolved is determined by self.__signature__
+        # The result is the initial ``scope`` the func nodes will both read from
+        # to get their arguments, and write their outputs to.
+        scope = self._get_kwargs(*args, **kwargs)
         # Go through self.func_nodes in order and call them on scope (performing said
         # read_input -> call_func -> write_output operations)
         self.call_on_scope(scope)
@@ -1097,6 +1145,105 @@ class DAG:
             dag = dag.add_edge(*edge)
         return dag
 
+    def ch_funcs(self, **name_and_func):
+        """
+        Change some of the functions in the DAG.
+        More preciseluy get a copy of the DAG where in some of the functions have
+        changed.
+
+        :param name_and_func: ``name=func`` pairs where ``name`` is the ``FuncNode.name``
+            of the func nodes you want to change and func is the function you want to
+            change it by.
+        :return: A new DAG with the different functions.
+
+        """
+
+        def ch_func(dag, name, func):
+            return DAG(
+                replace_item_in_iterable(
+                    dag.func_nodes,
+                    condition=lambda fn: fn.name == name,
+                    replacement=lambda fn: ch_func_node_attrs(fn, func=func),
+                )
+            )
+
+        new_dag = self
+        for name, func in name_and_func.items():
+            new_dag = ch_func(new_dag, name, func)
+        return new_dag
+
+    def debugger(self, feedback: Callable = dflt_debugger_feedback):
+        r"""
+        Utility to debug DAGs by computing each step sequentially, with feedback.
+
+        :param feedback: A callable that defines what feedback is given, usually used to
+            print/log some information and output some information for every step.
+            Must be a function with signature ``(func_node, scope, output, step)`` or
+            a subset thereof.
+        :return:
+
+        >>> from inspect import signature
+        >>>
+        >>> def f(a, b):
+        ...     return a + b
+        ...
+        >>> def g(c, d=4):
+        ...     return c * d
+        ...
+        >>> def h(f, g):
+        ...     return g - f
+        ...
+        >>> dag2 = DAG([f, g, h], name='arithmetic')
+        >>> dag2
+        DAG(func_nodes=[FuncNode(c,d -> g_ -> g), FuncNode(a,b -> f_ -> f), FuncNode(f,g -> h_ -> h)], name='arithmetic')
+        >>> str(signature(dag2))
+        '(c, a, b, d=4)'
+        >>> dag2(1,2,3)
+        -1
+        >>>
+        >>> debugger = dag2.debugger()
+        >>> str(signature(debugger))
+        '(c, a, b, d=4)'
+        >>> d = debugger(1,2,3)
+        >>> next(d)  # doctest: +NORMALIZE_WHITESPACE
+        0 --------------------------------------------------------------
+            func_node=FuncNode(c,d -> g_ -> g)
+            scope={'c': 1, 'a': 2, 'b': 3, 'd': 4, 'g': 4}
+        4
+        >>> next(d)  # doctest: +NORMALIZE_WHITESPACE
+        1 --------------------------------------------------------------
+            func_node=FuncNode(a,b -> f_ -> f)
+            scope={'c': 1, 'a': 2, 'b': 3, 'd': 4, 'g': 4, 'f': 5}
+        5
+
+        ... and so on. You can also choose to run every step all at once, collecting
+        the ``feedback`` outputs of each step in a list, like this:
+
+        >>> feedback_outputs = list(debugger(1,2,3))  # doctest: +NORMALIZE_WHITESPACE
+        0 --------------------------------------------------------------
+            func_node=FuncNode(c,d -> g_ -> g)
+            scope={'c': 1, 'a': 2, 'b': 3, 'd': 4, 'g': 4}
+        1 --------------------------------------------------------------
+            func_node=FuncNode(a,b -> f_ -> f)
+            scope={'c': 1, 'a': 2, 'b': 3, 'd': 4, 'g': 4, 'f': 5}
+        2 --------------------------------------------------------------
+            func_node=FuncNode(f,g -> h_ -> h)
+            scope={'c': 1, 'a': 2, 'b': 3, 'd': 4, 'g': 4, 'f': 5, 'h': -1}
+
+        """
+        # TODO: Add feedback callable validation
+        @Sig(self)
+        def launch_debugger(*args, **kwargs):
+            scope = self._get_kwargs(*args, **kwargs)
+            for step, func_node in enumerate(self.func_nodes):
+                output = func_node.call_on_scope(scope)
+                kwargs = dict(
+                    func_node=func_node, scope=scope, output=output, step=step
+                )
+                yield call_forgivingly(feedback, **kwargs)
+
+        return launch_debugger
+
     # ------------ display --------------------------------------------------------------
 
     def synopsis_string(self, bind_info='values'):
@@ -1105,7 +1252,15 @@ class DAG:
         )
 
     # TODO: Give more control (merge with lined)
-    def dot_digraph_body(self, start_lines=()):
+    def dot_digraph_body(
+        self,
+        start_lines=(),
+        *,
+        end_lines=(),
+        vnode_shape: str = dflt_configs['vnode_shape'],
+        fnode_shape: str = dflt_configs['fnode_shape'],
+        func_display: bool = dflt_configs['func_display'],
+    ):
         """Make lines for dot (graphviz) specification of DAG
 
         >>> def add(a, b=1): return a + b
@@ -1121,7 +1276,14 @@ class DAG:
         """
         if isinstance(start_lines, str):
             start_lines = start_lines.split()  # TODO: really? split on space?
-        yield from dot_lines_of_func_nodes(self.func_nodes, start_lines=start_lines)
+        if isinstance(end_lines, str):
+            end_lines = end_lines.split()
+        kwargs = dict(
+            vnode_shape=vnode_shape, fnode_shape=fnode_shape, func_display=func_display
+        )
+        yield from dot_lines_of_func_nodes(
+            self.func_nodes, start_lines=start_lines, end_lines=end_lines, **kwargs
+        )
 
     @wraps(dot_digraph_body)
     def dot_digraph_ascii(self, *args, **kwargs):
@@ -1155,14 +1317,6 @@ class DAG:
 
 # These are the defaults used in lined.
 # TODO: Merge some of the functionalities around graph displays in lined and meshed
-dflt_configs = dict(
-    fnode_shape='box',
-    vnode_shape='none',
-    display_all_arguments=True,
-    edge_kind='to_args_on_edge',
-    input_node=True,
-    output_node='output',
-)
 
 
 def param_to_dot_definition(p: Parameter, shape=dflt_configs['vnode_shape']):
@@ -1188,21 +1342,26 @@ def dot_lines_of_func_parameters(
     func_id: str,
     *,
     func_label: str = None,
-    output_shape: str = dflt_configs['vnode_shape'],
-    func_shape: str = dflt_configs['fnode_shape'],
+    vnode_shape: str = dflt_configs['vnode_shape'],
+    fnode_shape: str = dflt_configs['fnode_shape'],
+    func_display: bool = dflt_configs['func_display'],
 ) -> Iterable[str]:
     assert func_id != out, (
         f"Your func and output name shouldn't be the " f'same: {out=} {func_id=}'
     )
-    func_label = func_label or func_id
-    yield f'{out} [label="{out}" shape="{output_shape}"]'
-    yield f'{func_id} [label="{func_label}" shape="{func_shape}"]'
-    yield f'{func_id} -> {out}'
-    # args -> func
+    yield f'{out} [label="{out}" shape="{vnode_shape}"]'
     for p in parameters:
-        yield from param_to_dot_definition(p)
-    for p in parameters:
-        yield f'{p.name} -> {func_id}'
+        yield from param_to_dot_definition(p, shape=vnode_shape)
+
+    if func_display:
+        func_label = func_label or func_id
+        yield f'{func_id} [label="{func_label}" shape="{fnode_shape}"]'
+        yield f'{func_id} -> {out}'
+        for p in parameters:
+            yield f'{p.name} -> {func_id}'
+    else:
+        for p in parameters:
+            yield f'{p.name} -> {out}'
 
 
 def _parameters_and_names_from_sig(
@@ -1216,7 +1375,9 @@ def _parameters_and_names_from_sig(
     return sig.parameters, out, func_name
 
 
-def dot_lines_of_func_nodes(func_nodes: Iterable[FuncNode], start_lines=()):
+def dot_lines_of_func_nodes(
+    func_nodes: Iterable[FuncNode], start_lines=(), end_lines=(), **kwargs
+):
     r"""Got lines generator for the graphviz.DiGraph(body=list(...))
 
     >>> def add(a, b=1):
@@ -1302,12 +1463,14 @@ def dot_lines_of_func_nodes(func_nodes: Iterable[FuncNode], start_lines=()):
     yield from start_lines
     validate_that_func_node_names_are_sane(func_nodes)
     for func_node in func_nodes:
-        yield from dot_lines_of_func_node(func_node)
+        yield from dot_lines_of_func_node(func_node, **kwargs)
+    yield from end_lines
 
 
-def dot_lines_of_func_node(func_node: FuncNode):
+def dot_lines_of_func_node(func_node: FuncNode, **kwargs):
 
     out = func_node.out
+
     func_id = func_node.name
     func_label = getattr(func_node, 'func_label', func_id)
     if out == func_id:  # though forbidden in default FuncNode validation
@@ -1317,7 +1480,7 @@ def dot_lines_of_func_node(func_node: FuncNode):
     params = func_node.sig.ch_names(**func_node.bind).params
 
     yield from dot_lines_of_func_parameters(
-        params, out=out, func_id=func_id, func_label=func_label
+        params, out=out, func_id=func_id, func_label=func_label, **kwargs
     )
 
 
